import numpy as np
from numpy.linalg import norm


def pseudoweights_from_factor(w_factor: float, n: int, idx: int = 0):
    """Return pseudoweights array such that the first element is w_factor larger than the rest of the elements.
    These are pseudo-weights because the sum of all weights does not add up to 1."""
    w = np.ones(n, dtype="float")
    if n == 1:
        w[idx] == 1
        return w

    # Construct weight array with w_factor the factor the first component weighs vs the rest of the componentes
    w[idx] = w_factor
    return w


def weights_from_array(w):
    """Construct normalized weights array from an arbitrary array"""

    # Positive weights only
    w = np.abs(w)

    # Replace NaN with 0
    idx_nan = np.isnan(w)
    w[idx_nan] = 1

    # Array contains the same value
    if len(set(w)) == 1:
        return None

    # Normalize
    w /= norm(w, 1)

    # Normalization of very large / small values can also lead to the same value
    if len(set(w)) == 1:
        return None

    return w



"""Tests for nandist.pdist. Only checks for equivalence to scipy.spatial.pdist.
Numerical tests are implemented in test_cdist.py, because nandist.pdist calls nandist.cdist under the hood."""

import pytest
import scipy
import numpy as np
import nandist
from hypothesis import given
from .utils.weights import weights_from_array
from .strategies.arrays import normal_arrays


@pytest.mark.parametrize("metric", nandist.SUPPORTED_METRICS)
@given(X=normal_arrays())
def test_pdist_equivalent(X, metric):
    """Test whether the overloaded method `cdist` returns the same results as the original method when an array does not
    contain NaNs"""
    X, w = X[:2], X[2]

    w = weights_from_array(w)

    try:
        Y_expected = scipy.spatial.distance.pdist(X, metric=metric, w=w)
    except ZeroDivisionError:
        assert True
        return

    Y_result = nandist.pdist(X, metric=metric, w=w)
    assert np.array_equal(Y_result, Y_expected, equal_nan=True)



import pytest
import numpy as np
import scipy
import nandist
from hypothesis import given

from .strategies.arrays import normal_arrays, sized_normal_arrays, sized_nan_arrays
from .strategies.parameters import minkowski_p
from .utils.weights import weights_from_array


@pytest.mark.parametrize("metric", nandist.functions.functions)
@given(X=normal_arrays(), p=minkowski_p())
def test_equivalent_without_nans(X, metric, p):
    """Test whether nandist.<metric> is equivalent to scipy.spatial.distance.<metric> when passed normal arrays without
    NaNs. Asserts True if they return the same result or when they raise the same error"""

    # Unpack matrix with arrays
    u, v, w = X
    w = weights_from_array(w)

    nandist_f = getattr(nandist, metric)
    scipy_f = getattr(scipy.spatial.distance, metric)

    kwargs = {"u": u, "v": v, "w": w}

    # Add Minkowski's "p" parameter if the metric is "minkowski"
    if metric == "minkowski":
        kwargs = {**kwargs, "p": p}

    e_scipy = None
    e_nandist = None

    # Attempt to call the scipy.spatial.distance.<metric> function; store error if one is raised
    try:
        y_scipy = scipy_f(**kwargs)
    except Exception as e:
        e_scipy = e

    # Attempt to call the nandist.<metric> function; store error if one is raised
    try:
        y_nandist = nandist_f(**kwargs)
    except Exception as e:
        e_nandist = e

    # If an error was raised in either implementation; assert whether the same error was raised.
    # Return early
    if e_nandist is not None or e_scipy is not None:
        assert type(e_scipy) == type(e_nandist)
        return

    # If no error was raised but one of the results is NaN, assert whether both results are NaN.
    # Return early
    if np.isnan(y_scipy) or np.isnan(y_nandist):
        assert np.isnan(y_scipy) and np.isnan(y_nandist)
        return

    # In all other cases: assert whether the returned value is the same
    try:
        assert y_scipy == y_nandist
    except AssertionError:
        print("uh-oh")


@pytest.mark.parametrize("metric", nandist.functions.functions)
@given(X=sized_nan_arrays(2, 6), w=sized_normal_arrays(1, 6), p=minkowski_p())
def test_equivalent_with_nans(X, w, metric, p):
    """Test whether distance between u, v with weights w calculated with nandist.<metric> is equivalent to distance uh,
    vh with weights wh calculated with scipy.spatial.distance.<metric> when uh, vh the vectors u, v with components
    removed if they are NaN in either u or v. The weights vector wh is also adapted in a similar way."""
    u, v = X
    w = weights_from_array(w[0])

    nandist_f = getattr(nandist, metric)
    scipy_f = getattr(scipy.spatial.distance, metric)

    # Determine inputs to scipy
    idx_nan = np.isnan(u) | np.isnan(v)
    uh, vh = u[~idx_nan], v[~idx_nan]
    wh = w[~idx_nan] if w is not None else None

    kwargs_nandist = {"u": u, "v": v, "w": w}
    kwargs_scipy = {"u": uh, "v": vh, "w": wh}

    if metric == "minkowski":
        kwargs_nandist = {**kwargs_nandist, "p": p}
        kwargs_scipy = {**kwargs_scipy, "p": p}

    # Initialize errors
    e_nandist = None
    e_scipy = None

    try:
        y_nandist = nandist_f(**kwargs_nandist)
    except Exception as e:
        e_nandist = e

    try:
        y_scipy = scipy_f(**kwargs_scipy)
    except Exception as e:
        e_scipy = e

    # If an error was raised in either implementation; assert whether the same error was raised.
    # Return early
    if e_nandist is not None or e_scipy is not None:
        assert type(e_scipy) == type(e_nandist)
        return

    # If no error was raised but one of the results is NaN, assert whether both results are NaN.
    # Return early
    if np.isnan(y_scipy) or np.isnan(y_nandist):
        assert np.isnan(y_scipy) and np.isnan(y_nandist)
        return

    # In all other cases: assert whether the returned value is the same
    assert y_scipy == y_nandist



"""Test cdist functions. There is no need to test nandist.pdist because it calls cdist under the hood."""

from hypothesis.strategies import integers
from hypothesis import given
from .strategies.arrays import sized_normal_arrays, sized_nan_arrays
from .utils.weights import weights_from_array, pseudoweights_from_factor
from .strategies.parameters import minkowski_p
from .utils.constants import RELTOL, ZEROS_WITH_NAN, ONE_WITH_NAN

import pytest
import numpy as np
import scipy
import nandist

TEST_ARRAY_DIMENSION = 5


@pytest.mark.parametrize("metric", nandist.SUPPORTED_METRICS)
@given(
    X=sized_nan_arrays(number_of_arrays=8, array_dimension=TEST_ARRAY_DIMENSION),
    W=sized_normal_arrays(number_of_arrays=1, array_dimension=TEST_ARRAY_DIMENSION),
    p=minkowski_p(),
)
def test_cdist_correct_with_nans(X, W, p, metric):
    """Tests whehter nandist.cdist gives correct results when NaNs are present in the arrays.
    This leverages the nandist.<metric> function and should thus only pass when all tests for nandist.<metric>
    functions have passed"""
    XA, XB = X[:3], X[3:]
    XA_n, XB_n = XA.shape[0], XB.shape[0]

    w = weights_from_array(W[0])

    kwargs = {"metric": metric, "w": w}
    kwargs_nandist = {"w": w}

    if metric == "minkowski":
        kwargs = {**kwargs, "p": p}
        kwargs_nandist = {**kwargs_nandist, "p": p}

    e_nandist = None
    e_scipy = None

    try:
        Y_expected = np.zeros(shape=(XA_n, XB_n))
        nandist_f = getattr(nandist, metric)
        for i in range(XA_n):
            for j in range(XB_n):
                try:
                    y = nandist_f(XA[i], XB[j], **kwargs_nandist)
                    Y_expected[i, j] = y
                except ZeroDivisionError:
                    if metric == "cosine":
                        # Ignore it
                        Y_expected[i, j] = 0
    except Exception as e:
        e_scipy = e

    try:
        Y_result = nandist.cdist(XA, XB, **kwargs)
    except Exception as e:
        e_nandist = e

    if e_scipy is not None or e_nandist is not None:
        assert type(e_scipy) == type(e_nandist)
        return

    try:
        assert np.allclose(Y_result, Y_expected, atol=1e-6)
    except AssertionError:
        raise AssertionError


@pytest.mark.parametrize("a", ZEROS_WITH_NAN)
@pytest.mark.parametrize("b", ONE_WITH_NAN)
@pytest.mark.parametrize("p", [0.1, 0.5, 1, 2, 3, 4])
@given(w_factor=integers(min_value=1, max_value=100))
def test_weighted_minkowski(a, b, w_factor, p):
    """Test weighted minkowski assuming vector a is 1 at position 0 and zero elsewhere; b is zero everywhere except for
    positions where it is NaN"""

    XA, XB = np.array([a]), np.array([b])

    # Construct weights array
    w = pseudoweights_from_factor(w_factor=w_factor, n=XA.shape[1])

    # Calculate distance
    Y = nandist.cdist(XA, XB, "minkowski", w=w, p=p)

    assert (Y[0][0] - w[0] ** (1 / p)) / Y[0][0] < RELTOL


@pytest.mark.parametrize("metric", nandist.SUPPORTED_METRICS)
@given(
    X=sized_normal_arrays(number_of_arrays=8, array_dimension=TEST_ARRAY_DIMENSION),
    W=sized_normal_arrays(number_of_arrays=1, array_dimension=TEST_ARRAY_DIMENSION),
    p=minkowski_p(),
)
def test_cdist_equivalent_without_nans(X, W, p, metric):
    """Test whether nandist.cdist returns the same results as the scipy.spatial.distance.cdist when no NaNs are in the
    arrays"""
    XA, XB = X[:3], X[3:]

    w = weights_from_array(W[0])

    kwargs = {"metric": metric, "w": w}

    if metric == "minkowski":
        kwargs = {**kwargs, "p": p}

    try:
        Y_expected = scipy.spatial.distance.cdist(XA, XB, **kwargs)
    except ZeroDivisionError:
        assert True
        return

    Y_result = nandist.cdist(XA, XB, **kwargs)
    assert np.allclose(Y_result, Y_expected, equal_nan=True)



"""NA-ignoring replacement for scipy.spatial.distance.cdist"""
import itertools

from .functions import chebyshev as _chebyshev

import scipy
import numpy as np
from functools import wraps


def ignore_na_cdist(*args, **kwargs):
    @wraps(scipy.spatial.distance.cdist)
    def inner(XA, XB, metric="euclidean", *, out=None, **kwargs):
        """The arguments are taken from the scipy API"""
        # No NANs present? Do nothing
        nans_in_A, nans_in_B = np.any(np.isnan(XA)), np.any(np.isnan(XB))
        if not nans_in_A and not nans_in_B:
            return scipy.spatial.distance.cdist(XA, XB, metric, out=out, **kwargs)

        if metric == "cityblock":
            return minkowski(XA, XB, out=out, p=1, **kwargs)
        elif metric == "euclidean":
            return minkowski(XA, XB, out=out, p=2, **kwargs)
        elif metric == "minkowski":
            return minkowski(XA, XB, out=out, **kwargs)
        elif metric == "cosine":
            return cosine(XA, XB, out=out, **kwargs)
        elif metric == "chebyshev":
            return chebyshev(XA, XB, out=out, **kwargs)
        else:
            raise NotImplementedError(
                f"NA-ignored distance {metric =} not implemented."
            )

    return inner


@ignore_na_cdist
def cdist(*args, **kwargs):
    return


def minkowski(XA, XB, **kwargs):
    """NA-ignoring Minkowski variant returning a square form distance matrix"""
    if np.all(np.isnan(XA)) | np.all(np.isnan(XB)):
        return np.zeros(shape=(XA.shape[0], XB.shape[0]))

    p = kwargs.get("p", 2.0)  # get 'p' from kwargs, otherwise default to 2
    w = kwargs.get("w")
    out = kwargs.get("out")

    XA0 = np.nan_to_num(XA, nan=0)
    XB0 = np.nan_to_num(XB, nan=0)

    Dp = scipy.spatial.distance.cdist(XA0, XB0, "minkowski", out=out, w=w, p=p) ** p

    if w is None:
        w = np.ones_like(XA[0])

    Dp -= np.matmul(np.isnan(XA), (w * (np.abs(XB0) ** p)).T)
    Dp -= np.matmul(w * (np.abs(XA0) ** p), np.isnan(XB).T)

    # Hack: set very small values to zero to prevent inflation by ** (1/p) for larger values of p
    Dp[Dp / np.average(Dp) < 1e-10] = 0

    return np.nan_to_num(Dp ** (1 / p))


def zero_check(X, w=None):
    """Return whether the average weighted squared elements of vectors in X are zero"""
    return np.average(np.square(np.nan_to_num(X)), weights=w, axis=1, keepdims=1) == 0


def _correct_norms(XA, XB, w=None):
    """Correct norms ||u||*||v|| to use with cosine similarity normalizatation when XA or XB contain NaNs"""
    if w is None:
        w = np.ones(shape=(1, XA.shape[1]))
    else:
        w = np.array([w])
    return np.sqrt(
        np.matmul(np.square(np.nan_to_num(XA)), w.T * ~np.isnan(XB).T)
    ) * np.sqrt(np.dot(w * ~np.isnan(XA), np.square(np.nan_to_num(XB).T)))


def _uncorrected_norms(XA, XB, w=None):
    """Uncorrected norms ||u0||*||v0|| that were used with cosine similarity normalization when NaNs in XA and XB were
    zero-filled"""
    if w is None:
        w = np.ones(shape=(1, XA.shape[1]))
    return np.sqrt(
        np.matmul(np.square(np.nan_to_num(XA)), (w * np.ones_like(XB)).T)
    ) * np.sqrt(np.dot(w * np.ones_like(XA), np.square(np.nan_to_num(XB)).T))


def cosine(XA, XB, **kwargs):
    """NA-ignoring cosine distance variant returning a square form distance matrix"""

    XA0 = np.nan_to_num(XA, nan=0)
    XB0 = np.nan_to_num(XB, nan=0)

    # Cosine similarity calculated with XA0 and XB0 are incorrect due to an inflation of the normalization factor
    # We must correct for this factor
    w = kwargs.get("w", None)

    uuvv = _correct_norms(XA, XB, w=w)
    uu0vv0 = _uncorrected_norms(XA, XB, w=w)

    # Cosine may return nan when the vectors are all zeros, so convert before returning
    D = np.nan_to_num(scipy.spatial.distance.cdist(XA0, XB0, "cosine", **kwargs))

    # Corrections may cause division by zero, so also convert NaNs to zero in corrected distance
    return np.nan_to_num(1 - (uu0vv0 / uuvv) * (1 - D))


def chebyshev(XA, XB, **kwargs):
    """NA-ignoring chebyshev distance variant returning a square form distance matrix"""

    # Unfortunately, we need to loop because the Chebyshev distance takes the maximum distance and hence there is no
    # algebraic way to correct for overestimations when NaNs are replaced with zero.
    Y = np.zeros(shape=(XA.shape[0], XB.shape[0]))
    i = -1
    for x, (u, v) in enumerate(itertools.product(XA, XB)):
        # Call nandist.chebyshev
        y = _chebyshev(u, v, w=kwargs.get("w", None))

        # Index to distance matrix
        j = x % XB.shape[0]
        # Increment
        if j == 0:
            i += 1

        Y[i, j] = y

    return Y



"""Compute distances in numpy arrays with nans"""
__version__ = "0.12.0"

from .cdist import cdist  # noqa: F401
from .pdist import pdist  # noqa: F401
from .functions import (
    braycurtis,  # noqa: F401
    canberra,  # noqa: F401
    chebyshev,  # noqa: F401
    cityblock,  # noqa: F401
    correlation,  # noqa: F401
    cosine,  # noqa: F401
    euclidean,  # noqa: F401
    jensenshannon,  # noqa: F401
    mahalanobis,  # noqa: F401
    minkowski,  # noqa: F401
    seuclidean,  # noqa: F401
    sqeuclidean,  # noqa: F401
)

SUPPORTED_METRICS = [
    "chebyshev",
    "cityblock",
    "cosine",
    "euclidean",
    "minkowski",
]



from hypothesis.strategies import composite, floats, integers


@composite
def minkowski_p(draw):
    """Draw minkowski p, which can be an integer between 1 and 5 (typical usage) or any float between zero and 1"""
    return draw(
        floats(min_value=0.01, max_value=1, exclude_min=True, exclude_max=True)
        | integers(min_value=1, max_value=5)
    )



from hypothesis.extra.numpy import arrays, from_dtype
from hypothesis.strategies import integers, composite, booleans
import numpy as np


# Array lengths
MIN_LENGTH = 1
MAX_LENGTH = 1_000

# We test using 32-bit floats because testing with 64-bit floats can lead to infinities and accumulation of inaccuracies
# when extreme values (like the maximum float) are involved; breaking the tests.
# The goal of the tests is to test the validity of the corrections, and not the robustness of nandist for extreme inputs
TEST_DTYPE = "int32"


@composite
def array_elements(draw, nan: bool = True):
    """Strategy for generating values to be used in testing. These are integer values between -10 and 10 and can
    sometimes be nan"""

    # Are we returning a NaN?
    if nan:
        if draw(booleans()):
            return np.nan

    i = draw(integers(min_value=-10, max_value=10))
    return float(i)


@composite
def normal_arrays(draw, min_length=MIN_LENGTH, max_length=MAX_LENGTH):
    """Hypothesis strategy for returning a matrix with shape = (3, length), with random length between min_length and
    max_length"""
    length = draw(integers(min_value=min_length, max_value=max_length))

    # Draw random array, ensure that the sum of elements is less than the max float to prevent testing errors
    return draw(
        arrays(
            np.dtype("float"),
            shape=(3, length),
            elements=from_dtype(
                np.dtype(TEST_DTYPE), allow_nan=False, allow_infinity=False
            ),
        ).filter(lambda x: np.sum(x) < np.finfo("float").max)
    )


@composite
def nan_arrays(draw, min_length=MIN_LENGTH, max_length=MAX_LENGTH):
    """Hypothesis strategy for returning a matrix with shape = (3, length), with random length between min_length and
    max_length. The matrix consists of vectors a, b, and w.
    The filter ensures that either a or b contains at least one NAN value"""
    length = draw(integers(min_value=min_length, max_value=max_length))

    # Draw random array, filter such that:
    # 1: there is at least one NaN in the first two arrays
    # 2: the sum of all elements is smaller than the maximum float

    # The reason for (2) is that we will be testing whether the nandist calculated distance is equal to the scipy
    # distance alculated for an array where NaN elements are removed. In our implementation, we might overestimate the
    # distance and then correct for this overestimation. However, the overestimation can lead to infinite values.
    # To prevent these infinities, the sum of elements of each vector cannot exceed the maximum float value.
    return draw(
        arrays(
            np.dtype("float"),
            shape=(3, length),
            elements=from_dtype(
                np.dtype(TEST_DTYPE),
                allow_nan=True,
                allow_infinity=False,
            ),
        )
        .filter(lambda x: np.any(np.isnan(x[:2])))  # filter (1)
        .filter(
            lambda x: np.all(np.sum(np.nan_to_num(x, 0), 1) < np.finfo("float").max)
        )  # filter (2)
    )


@composite
def sized_normal_arrays(draw, number_of_arrays: int, array_dimension: int):
    """Return a NaN-free matrix with shape = (number_of_arrays, array_dimension)"""
    # Draw random array, ensure that the sum of elements is less than the max float to prevent testing errors
    return draw(
        arrays(
            np.dtype("float"),
            shape=(number_of_arrays, array_dimension),
            elements=array_elements(nan=False),
        )
        .filter(lambda x: np.sum(x) < np.finfo("float").max)
        .filter(lambda x: np.any(np.sum(x) > 0))
    )


@composite
def sized_nan_arrays(draw, number_of_arrays: int, array_dimension: int):
    """Return a NaN-containing matrix with shape (number_of_arrays, array_dimensions)"""
    return draw(
        arrays(
            np.dtype("float"),
            shape=(number_of_arrays, array_dimension),
            elements=array_elements(nan=True),
        ).filter(lambda x: np.any(np.isnan(x)))
    )



"""NA-ignoring distance functions for calculating the distance between two arrays"""

from functools import wraps
import numpy as np
import scipy

functions = [
    "braycurtis",
    "canberra",
    "chebyshev",
    "cityblock",
    "correlation",
    "cosine",
    "euclidean",
    "jensenshannon",
    "mahalanobis",
    "minkowski",
    "seuclidean",
    "sqeuclidean",
]


def ignore_na(f):
    """Ignore NaNs in input vectors to distance function f"""
    scipy_f = getattr(scipy.spatial.distance, f.__name__)

    @wraps(scipy_f)
    def inner(u, v, w=None, **kwargs):
        """Ignore the NaNs in input vectors u and v and optionally in w"""
        na_u = np.isnan(u)
        na_v = np.isnan(v)

        na_uv = na_u | na_v

        # if np.all(na_uv):
        #    return 0

        if w is not None:
            w = w[~na_uv]

        return scipy_f(u[~na_uv], v[~na_uv], w=w, **kwargs)

    return inner


@ignore_na
def braycurtis(*args, **kwargs):
    return


@ignore_na
def canberra(*args, **kwargs):
    return


@ignore_na
def chebyshev(u, v, w=None):
    return


@ignore_na
def cityblock(*args, **kwargs):
    return


@ignore_na
def correlation(*args, **kwargs):
    return


@ignore_na
def cosine(*args, **kwargs):
    return


@ignore_na
def euclidean(*args, **kwargs):
    return


@ignore_na
def jensenshannon(*args, **kwargs):
    return


@ignore_na
def mahalanobis(*args, **kwargs):
    return


@ignore_na
def minkowski(*args, **kwargs):
    return


@ignore_na
def seuclidean(*args, **kwargs):
    return


@ignore_na
def sqeuclidean(*args, **kwargs):
    return



"""NA-ignoring replacement for scipy.spatial.distance.pdist
Calls cdist implementation and then converts to appropriate distance matrix form"""

import scipy
import numpy as np
from functools import wraps

from .cdist import minkowski, chebyshev, cosine


def convert_form(Y):
    """Return converted form of distance matrix. Required to convert cdist result form to expected pdist result form"""
    return scipy.spatial.distance.squareform(Y)


def ignore_na_pdist(*args, **kwargs):
    @wraps(scipy.spatial.distance.pdist)
    def inner(X, metric="euclidean", *, out=None, **kwargs):
        """The arguments are taken from the scipy API"""

        # No NANs present? Return scipy.spatial.distance.pdist result
        if not np.any(np.isnan(X)):
            return scipy.spatial.distance.pdist(X, metric, out=out, **kwargs)

        if metric == "cityblock":
            return convert_form(minkowski(X, X, out=out, p=1, **kwargs))
        elif metric == "euclidean":
            return convert_form(minkowski(X, X, out=out, p=2, **kwargs))
        elif metric == "cosine":
            return convert_form(cosine(X, X, out=out, **kwargs))
        elif metric == "chebyshev":
            return convert_form(chebyshev(X, X, out=out, **kwargs))
        else:
            raise NotImplementedError(
                f"NA-ignored distance {metric =} not implemented."
            )

    return inner


@ignore_na_pdist
def pdist(*args, **kwargs):
    return



import numpy as np

EPS = np.finfo("float").eps
RELTOL = 1e-12


ZEROS_WITH_NAN = [
    np.array([0, 0, 0]),
    np.array([0, np.NAN, 0]),
    np.array([0, 0, np.NAN]),
    np.array([0, np.NAN, np.NAN]),
]

ONE_WITH_NAN = [
    np.array([1, 0, 0]),
    np.array([1, np.NAN, 0]),
    np.array([1, np.NAN, np.NAN]),
]
